/* Empfänger_demo_2
  zyklische Übertragung einiger Testdaten mit Zeitsteuerung
*/

// Bibliotheken
#include <limits.h>
#include <DataReceiver.h>
#include <util/atomic.h>

//#######################################################################################################################

// einige Variablen & Konstanten

const uint8_t BYTE_RATE = 30; // die Übertragungsgeschwindigkeit (4...255 Bytes/s)
const uint8_t LOOKUP_TABLE_START = 0; // EEPROM Startadresse zur Speicherung des 256 Byte CRC-lookup table
const uint8_t RECEIVER_PIN = 2; // das (interruptfähige, digitale) Pin, an welchem die Daten eingelesen werden

dset_r dataset[2]; // ein Array mit zwei Elementen von Typ dset_r, welche jeweils einen Datensatz charakterisieren

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Dieser Teil ist neu!!! / Variablen, welche die Zeitsteuerung benötigt

// die Ziel-Schleifenlaufzeit für "loop" in µs (max. 16383)
const uint16_t TLR = 500; // 500µs = 0.5ms

// Schleifenwert für die Ausführung von Aufgabe 0 (1000 Schleifen x 0.5ms = 500ms)
const uint16_t TASK_0_EXEC = 1000;
// der zugehörige Schleifenzähler
uint16_t task_0_exec_cnt = 0;

// nach dem gleichen Muster können hier weitere "Aufgaben" definiert werden
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//#######################################################################################################################

// dataset[0] ist verfügbar -> prozessiere Daten
// die Prozessierung kurz halten, innerhalb dieser Funktion werden keine neuen Daten empfangen!!!
void processDataset_0() {
  Serial.print(F("Er wurde bei folgendem Zeitstempel empfangen: "));Serial.println(DR.getTimestamp());
  // Daten auslesen
  float Pi;
  uint32_t sys_time;
  DR.dataTransfer(&Pi, 4, 0, 0);
  DR.dataTransfer(&sys_time, 4, 4, 0);
  // Daten ausdrucken (über USB an den PC ausgeben)
  Serial.print(F("Der Datensatz enthält folgenden Wert für PI: "));Serial.println(Pi, 5);
  Serial.print(F("Außerdem wurde die Systemzeit des Senders übermittelt: "));Serial.println(sys_time);
  Serial.println(F("------------------------------------------------------------------------------------------------"));
}

//#######################################################################################################################

// dataset[1] ist verfügbar -> prozessiere Daten
// die Prozessierung kurz halten, innerhalb dieser Funktion werden keine neuen Daten empfangen!!!
void processDataset_1() {
  Serial.print(F("Er wurde bei folgendem Zeitstempel empfangen: "));Serial.println(DR.getTimestamp());
  // Daten auslesen
  char test_string[40];
  DR.dataTransfer(test_string, sizeof(test_string), 0, 0);
  // Daten ausdrucken (über USB an den PC ausgeben)
  Serial.print(F("Der Datensatz enthält eine Zeichenkette mit dem Inhalt: "));Serial.println(test_string);
  Serial.println(F("------------------------------------------------------------------------------------------------"));
}

//#######################################################################################################################

void setup() {

/*
Initialisierung der CRC8-Bibliothek (wird durch DataReceiver.h automatisch mit eingebunden)

Parameter:
 - EPROM Startadresse zur Speicherung des 256 Byte CRC-lookup table

 Rückgabe: die nächste freie EEPROM-Adresse
*/
  CRC.init(LOOKUP_TABLE_START);

/* 
erzeuge zwei Datensätze
dataset 0 hat die ID 10, eine Größe von 8 Byte und ist verbunden mit der Funktion "processDataset_0"
dataset 1 hat die ID 11, eine Größe von 40 Byte und ist verbunden mit der Funktion "processDataset_1"
*/
  dataset[0] = DR.createDataset(10, 8, processDataset_0);
  dataset[1] = DR.createDataset(11, 40, processDataset_1);

/*
Initialisierung der DataReceiver-Bibliothek

Parameter DataReceiver (v.l.n.r.):
 - Datenrate in Byte/s
 - das (interruptfähige, digitale) Pin, an welchem die Daten eingelesen werden
 - das "dset_r"-Array, welches alle relevanten Informationen zu den zu emfangenden Datensätzen enthält
 - die Anzahl der Datensätze (also die Zahl der Elemente von "dataset")
*/
  if(!DR.init(BYTE_RATE, RECEIVER_PIN, dataset, 2)) {
    pinMode(13, OUTPUT); // wenn die Initialisierung schief geht, hält der Code hier an und die on-board LED blinkt
    while(true) {
      digitalWrite(13, !digitalRead(13));
      delay(200);
      digitalWrite(13, !digitalRead(13));
      delay(200);
    }
  }

  Serial.begin(9600); // initialisiere die Schnittstelle zum PC, um die empfangenen Daten auszugeben
  Serial.println(F("Schnittstelle ist für die Datenausgabe bereit.\n"));

}

//#######################################################################################################################

void loop() {

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// wird für die Zeitsteuerung benötigt, nicht verändern
  static uint32_t stp = micros();
  static uint32_t st;
  static int32_t dt = 0;
  const int32_t MIN_DT = LONG_MIN>>1;
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

// Aufgabe 0 ausführen, wenn die Zahl der Schleifendurchläufe erreicht ist (aller 500ms)
  if(TASK_0_EXEC == task_0_exec_cnt) {
    // ~~~~~~~ Begin Aufgaben erledigen ~~~~~~~~~~~~~

    ATOMIC_BLOCK(ATOMIC_RESTORESTATE) { // dieser Block stellt sicher, das die Ausführung innerhalb nicht unterbrochen wird 
      if(DR.getStatus()&4) {//prüfe, ob Daten vorliegen und falls ja, ...
        DR.setStatus(2); // ... markiere den Datensatz als "busy reading" und ...
        if(DR.validateData()) { // ... prüfe, ob sie konsistent sind und ...
          Serial.print(F("Habe einen Datensatz empfangen. Dieser hat die ID: "));Serial.println(DR.getID());
          (dataset + DR.getPos())->FUNCTION(); //... prozessiere die Daten durch Aufruf der zugehörigen Funktion
        }
        DR.setStatus(0); // gebe den Datensatz wieder frei, um den Empfang weiterer Daten zu ermöglichen
      }
    }
    
    // ~~~~~~~ Ende Aufgaben erledigen ~~~~~~~~~~~~~
    
    task_0_exec_cnt = 0; // den Schleifenzähler auf 0 stellen
  }

// alle Schleifenzähler erhöhen
  ++task_0_exec_cnt;

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Steuerung der Schleifenlaufzeit / nicht verändern
  
// check if current system time is bigger then the timestamp obtained at the last loop
  dt += TLR;
  st = micros();
  if(st > stp) { dt -= int32_t(st - stp); }
  else { dt -= int32_t((ULONG_MAX - stp) + st + 1); }
  stp = st;
  
// if runtime-budget is larger then target, insert delay-step
  if(dt > TLR) {
    delayMicroseconds(dt);
    stp += dt;
    dt = 0;
  }
  else {
// avoid overflow, in case the budget is constantly negative (loop is too slow to meet target loop time)
    if(dt < MIN_DT) { dt = 0; }
  }
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

}
